#pragma once
#include "simpleCL.h"	// The SimpleOpenCL library
#include <iostream>		// For input/output
#include <stdlib.h>		// For random function
#include <time.h>		// For random function initialization
#include <string>		// For string functions
#include "Utils.h"
using namespace std;

const bool printFullArray = true;		// Indicate if the full output arrays have to be printed
// Exclusive scan tester
Result executeSimple2D(int inputSize, int *input) {
	unsigned int height = ceil(log2f(inputSize))+1;
	int paddedInputSize = pow(2, (height-1));
	size_t bytesGPU = paddedInputSize*sizeof(int);
	Result result;

	// Arrays
	int *inputGPU = padInputTo(input, inputSize, paddedInputSize);
	int *tempGPU = new int[paddedInputSize*height];
	int *outputGPU = new int[paddedInputSize];

	debug("Using input size of %d values, generating input values...\n", inputSize);

	// Print hardware info
	char* value = getHardwareName(getHardware());
	debug("Using hardware device: %s\n", value);

	// Determine maximum work group size and print it
	size_t workGroupSize = getMaximumWorkgroupSize(getHardware());
	debug("Using maximum workgroup size of %d.\n", workGroupSize);

	// Get correct kernels
	sclSoft software = sclGetCLSoftware("GPUSimple2D.cl", "prefixSum", getHardware());

	// Setup global and local sizes
	size_t globalSize[2], localSize[2];
	globalSize[0] = paddedInputSize/2;
	globalSize[1] = 1;
	localSize[0] = paddedInputSize/2;
	localSize[1] = 1;
	debug("  localSize=[%d, %d]\n", localSize[0], localSize[1]);
	debug("  globalSize=[%d, %d]\n", globalSize[0], globalSize[1]);

	// Launch kernel with correct arguments
	cl_event kernelResult = NULL;
	if (printFullArray) {   // Only when debugging the temporary array will be accesible from the host
		kernelResult = sclManageArgsLaunchKernel(getHardware(), software, globalSize, localSize,
			" %r %w %R %a %a ",
			bytesGPU, inputGPU,		// Input array
			bytesGPU, outputGPU,	// Output array
			bytesGPU * height, tempGPU,		// Temporary array for the tree
			sizeof(unsigned int), (void*)&paddedInputSize,
			sizeof(unsigned int), (void*)&height
			);
	}
	else {
		kernelResult = sclManageArgsLaunchKernel(getHardware(), software, globalSize, localSize,
			" %r %w %g %a %a ",
			bytesGPU, inputGPU,		// Input array
			bytesGPU, outputGPU,	// Output array
			bytesGPU * height,				// Temporary array for the tree
			sizeof(unsigned int), (void*)&paddedInputSize,
			sizeof(unsigned int), (void*)&height
			);
	}

	// Calculate the result with the CPU
	debug("GPU calculation complete, CPU calculation started.\n");
	int *outputCPU = calculateCPU(inputSize, input);

	// Display complete CPU and GPU arrays
	if (printFullArray) {
		printArrays(inputSize, input, outputCPU, outputGPU);
		debug("\n");
		printMultiArray(paddedInputSize, height, tempGPU); // Does not work with multidimensional arrays
	}

	// Compare the results of the CPU with the GPU
	int incorrect = 0;
	for (int i = 0; i < inputSize; i++) {
		if (outputCPU[i] != outputGPU[i]) {
			incorrect++;
		}
	}
	if (incorrect == 0) {
		debug("\nCorrect result, all values match.\n");
	}
	else {
		debug("\nFAILED, %d incorrect results out of %d output values.\n", incorrect, inputSize);
	}

	// Print the time it took to run the program
	cl_ulong time = sclGetEventTime(getHardware(), kernelResult);
	result.executionTime = time;
	debug("\nExecution time: %lu nanoseconds, %0.3f milliseconds\n", time, time / 1000000.0);
	result.output = outputGPU;

	// Release host memory
	delete[] inputGPU;
	delete[] input;
	delete[] tempGPU;
	delete[] outputCPU;

	return result;
}